﻿Public Class UmelaInteligence

#Region "Inherits"
    Inherits HerniDeska
#End Region

#Region "Constants"
    ''' <summary>
    ''' Hodnota udávající výhru či prohru.
    ''' </summary>
    ''' <remarks></remarks>
    Private Const MAT As Single = 1.0E+35
    Private Const MAT_instant As Single = 1.0E+38

    ''' <summary>
    ''' Hodnota udávající pozici blízko výhry či prohry.
    ''' </summary>
    ''' <remarks></remarks>
    Private Const MNOHO As Single = 1.0E+32

    ''' <summary>
    ''' Hodnota reprezentující mínus nekonečno.
    ''' </summary>
    ''' <remarks></remarks>
    Private Const MinusNekonecno As Single = -3.4028235E+38

    ''' <summary>
    ''' Hodnota pro pětku která je obsazena oběma hráči.
    ''' </summary>
    ''' <remarks></remarks>
    Private Const BN As Single = 0

    ''' <summary>
    ''' Hodnota pro prázdnou pětku.
    ''' </summary>
    ''' <remarks></remarks>
    Private Const B0 As Single = 0.5

    ''' <summary>
    ''' Hodnota pro pětku obsazenou 1 hráčem s 1 symbolem.
    ''' </summary>
    ''' <remarks></remarks>
    Private Const B1 As Single = 1.5

    ''' <summary>
    ''' Hodnota pro pětku obsazenou 1 hráčem s 2 symboly.
    ''' </summary>
    ''' <remarks></remarks>
    Private Const B2 As Single = 10.5

    ''' <summary>
    ''' Hodnota pro pětku obsazenou 1 hráčem s 3 symboly.
    ''' </summary>
    ''' <remarks></remarks>
    Private Const B3 As Single = 40.5

    ''' <summary>
    '''  Hodnota pro pětku obsazenou 1 hráčem s 4 symboly.
    ''' </summary>
    ''' <remarks></remarks>
    Private Const B4 As Single = 120.5

    ''' <summary>
    '''  Hodnota pro pětku obsazenou 1 hráčem s 5 symboly -> MAT.
    ''' </summary>
    ''' <remarks></remarks>
    Private Const B5 As Single = 1.0E+35

    ''' <summary>
    ''' Proměnná určující násobící faktor bonusu.
    ''' </summary>
    ''' <remarks></remarks>
    Private Const BM As Single = 25.0
#End Region

#Region "Varibles"
    ''' <summary>
    ''' Proměnná pro uložení Bonusu Pole.
    ''' </summary>
    ''' <remarks></remarks>
    Private FieldBonus() As Single = {0.001 * BM, 0.01 * BM, 0.001 * BM, 0.001 * BM, 0.01 * BM, 0.001 * BM, 0.01 * BM, 0.1 * BM, 0.01 * BM, 0.01 * BM, 0.1 * BM, 0.01 * BM, 0.001 * BM, 0.01 * BM, 0.001 * BM, 0.001 * BM, 0.01 * BM, 0.001 * BM, 0.001 * BM, 0.01 * BM, 0.001 * BM, 0.001 * BM, 0.01 * BM, 0.001 * BM, 0.01 * BM, 0.1 * BM, 0.01 * BM, 0.01 * BM, 0.1 * BM, 0.01 * BM, 0.001 * BM, 0.01 * BM, 0.001 * BM, 0.001 * BM, 0.01 * BM, 0.001 * BM}
#End Region



    ''' <summary>
    ''' Procedura na převzetí aktuálního stavu dané hry.
    ''' </summary>
    ''' <param name="HerniDeska">Herní deska k převzatí stavu</param>
    ''' <remarks></remarks>
    Public Sub PrevezmiStavHry(ByVal HerniDeska As HerniDeska)

        _BilyCekaNaVyhru = HerniDeska._BilyCekaNaVyhru
        _BilyRotoval = HerniDeska._BilyRotoval
        _BilyTahnul = HerniDeska._BilyTahnul
        _CernyCekaNaVyhru = HerniDeska._CernyCekaNaVyhru
        _CernyRotoval = HerniDeska._CernyRotoval
        _CernyTahnul = HerniDeska._CernyTahnul
        _CisloTahu = HerniDeska._CisloTahu
        _HracBily = HerniDeska._HracBily
        _HracCerny = HerniDeska._HracCerny
        _HracNaTahu = HerniDeska._HracNaTahu
        _Pole = HerniDeska._Pole
        _PosledniTah = HerniDeska._PosledniTah

    End Sub

    ''' <summary>
    ''' Funkce porovnávající pozice navzajem.
    ''' </summary>
    ''' <param name="Pozice1">Array(35) of SByte</param>
    ''' <param name="Pozice2">Array(35) of SByte</param>
    ''' <returns></returns>
    ''' <remarks></remarks>
    Private Function PorovnejPozice(ByVal Pozice1() As SByte, ByVal Pozice2() As SByte) As Boolean

        For a As Byte = 0 To 35
            If Pozice1(a) <> Pozice2(a) Then
                Return False
                Exit Function
            End If
        Next

        Return True

    End Function

    ''' <summary>
    ''' Funkce porovnávající tahy, zda nejsou navzájem symetrické.
    ''' </summary>
    ''' <param name="Tah1">Value of Short</param>
    ''' <param name="Tah2">Value of Short</param>
    ''' <returns></returns>
    ''' <remarks></remarks>
    Public Function PorovnejTahy(ByVal Tah1 As Short, ByVal Tah2 As Short) As Boolean

        Dim PoziceTahu1(35) As SByte
        Dim PoziceTahu2(35) As SByte

        For a As Byte = 0 To 35
            PoziceTahu1(a) = _Pole(a)
            PoziceTahu2(a) = _Pole(a)
        Next

        Dim Tah1_Rotace As Byte = Tah1 \ 36
        Dim Tah2_Rotace As Byte = Tah2 \ 36

        Dim Tah1_Tah As Byte = Tah1 - 36 * Tah1_Rotace
        Dim Tah2_Tah As Byte = Tah2 - 36 * Tah2_Rotace

        AI_ZahrajTah_SP(PoziceTahu1, Tah1_Tah, Tah1_Rotace, 1)
        AI_ZahrajTah_SP(PoziceTahu2, Tah2_Tah, Tah2_Rotace, 1)

        For a As Byte = 0 To 3
            If PorovnejPozice(PoziceTahu1, PoziceTahu2) = True Then
                Return True
                Exit For
            End If
            RotujPoziciDoprava(PoziceTahu2)
        Next

        Return False

    End Function

    ''' <summary>
    ''' Procedura rotující celou pozici doprava.
    ''' </summary>
    ''' <param name="Pozice">Array(35) of SByte</param>
    ''' <remarks></remarks>
    Private Sub RotujPoziciDoprava(ByRef Pozice() As SByte)

        Dim Backup(35) As SByte
        BackUp_Board(Pozice, Backup)

        For b As Byte = 0 To 5
            For a As Byte = 0 To 5
                Pozice(b + a * 6) = Backup((30 - b * 6) + a)
            Next
        Next

    End Sub

    ''' <summary>
    ''' Procedura redukující po prevedení a přerotování stejné tahy.
    ''' </summary>
    ''' <param name="Tahy">Array(PocetMoznychTahu) of Short</param>
    ''' <remarks></remarks>
    Private Sub RedukujTahy(ByRef Tahy() As Short)

        Dim NoveTahy(Tahy.Length - 1) As Short
        Dim NoveTahy_i As Integer
        Dim Hledani_i As Integer
        Dim i As Integer = 1
        Do

            If Tahy(Hledani_i) <> 32767 Then

                NoveTahy(NoveTahy_i) = Tahy(Hledani_i)
                NoveTahy_i += 1

                For a As Integer = 1 To Tahy.Length - 1 - Hledani_i
                    If Tahy(Hledani_i + a) <> 32767 Then
                        If PorovnejTahy(Tahy(Hledani_i), Tahy(Hledani_i + a)) = True Then

                            Tahy(Hledani_i + a) = 32767

                        End If
                    End If
                Next

            End If

            Hledani_i += 1

            If Hledani_i = Tahy.Length - 1 Then Exit Do

        Loop

        Do

            If NoveTahy(i) = 0 Then Exit Do
            i += 1

        Loop

        i -= 1

        ReDim Tahy(i)

        For a = 0 To i
            Tahy(a) = NoveTahy(a)
        Next

    End Sub




    ''''MOZNA OPTIMALIZACE NA COLLECTION OF SHORT

    ''' <summary>
    ''' Funce která vrací počet možných tahůz určené pozice.
    ''' </summary>
    ''' <returns></returns>
    ''' <remarks></remarks>
    Private Function PocitejMozneTahy() As Short

        Dim MoznychTahu As Short

        For a = 0 To 35
            If _Pole(a) = 0 Then MoznychTahu += 8
        Next

        MoznychTahu -= 1

        Return MoznychTahu

    End Function

    ''' <summary>
    ''' Procedura která vrací všechny možné tahy z určité pozice.
    ''' </summary>
    ''' <param name="Tahy">Array(PocetMoznychTahu) of Short</param>
    ''' <remarks></remarks>
    Private Sub GenerujMozneTahy(ByRef Tahy() As Short)

        Dim i As Short = 0

        For a = 0 To 35
            If _Pole(a) = 0 Then
                For b As Byte = 0 To 7

                    Tahy(i) = a + 36 * b
                    i += 1

                Next
            End If
        Next

    End Sub

    ''''/MOZNA OPTIMALIZACE NA COLLECTION OF SHORT


    ''' <summary>
    ''' Benchmarkovací funkce zjišťující jak rychle je schopen daný engine počítat.
    ''' </summary>
    ''' <param name="Engine123">Value of Byte</param>
    ''' <remarks></remarks>
    Public Sub Benchmark(ByVal Engine123 As Byte)

        '''''''''''''''''''''''''''''''hm'''
        UInteligence.PrevezmiStavHry(Deska)
        '''''''''''''''''''''''''''''''/hm''

        PocetPrubehuHodnoticiFunkce = 0

        Dim a As Short

        Dim StartTime As DateTime = Now

        Select Case Engine123
            Case 1
                a = NaivniAI_NejlepsiTah(_HracNaTahu, 0)
            Case 2
                a = MinMaxAI_NejlepsiTah(2, _HracNaTahu)
            Case 3
                a = AlfaBetaAI_NejlepsiTah(2, _HracNaTahu)
        End Select

        Dim Time As TimeSpan = Now - StartTime

        Alert_Benchmark(PocetPrubehuHodnoticiFunkce, Time)

        PocetPrubehuHodnoticiFunkce = 0

    End Sub


    ''' <summary>
    ''' Hodnotící funkce je jádrem každé AI.
    ''' </summary>
    ''' <param name="HracNaTahu">Value of SByte</param>
    ''' <param name="PosledniProvedenyTah">Value of Byte</param>
    ''' <returns></returns>
    ''' <remarks></remarks>
    Private Function HodnotaPozice(ByVal HracNaTahu As SByte, ByVal PosledniProvedenyTah As Byte)

        PocetPrubehuHodnoticiFunkce += 1

        Dim v_HodnotaPozice As Single

        'Získej všecky možný pětky
        Dim Petky(31, 4) As SByte
        Dim i As Short = 0

        For c = 0 To 1
            For b = 0 To 5
                For a = 0 To 4
                    Petky(i, a) = _Pole(a + c + b * 6)
                Next
                i += 1
            Next
        Next

        For c = 0 To 1
            For b = 0 To 5
                For a = 0 To 4
                    Petky(i, a) = _Pole(b + (a + c) * 6)
                Next
                i += 1
            Next
        Next

        For b = 4 To 5
            For a = 0 To 4
                Petky(i, a) = _Pole(b + 5 * a)
            Next
            i += 1
        Next

        For b = 10 To 11
            For a = 0 To 4
                Petky(i, a) = _Pole(b + 5 * a)
            Next
            i += 1
        Next

        For b = 0 To 1
            For a = 0 To 4
                Petky(i, a) = _Pole(b + 7 * a)
            Next
            i += 1
        Next

        For b = 6 To 7
            For a = 0 To 4
                Petky(i, a) = _Pole(b + 7 * a)
            Next
            i += 1
        Next

        Dim pocet_plus1, pocet_minus1 As Byte

        For a = 0 To 31

            pocet_plus1 = 0
            pocet_minus1 = 0

            For b = 0 To 4
                Select Case Petky(a, b)
                    Case 1
                        pocet_plus1 += 1
                    Case -1
                        pocet_minus1 += 1
                End Select
            Next

            If pocet_minus1 > 0 And pocet_plus1 > 0 Then
                v_HodnotaPozice += BN
            ElseIf pocet_minus1 = 0 And pocet_plus1 = 0 Then
                v_HodnotaPozice += B0
            ElseIf pocet_minus1 = 0 Then
                Select Case pocet_plus1
                    Case 1
                        v_HodnotaPozice += B1
                    Case 2
                        v_HodnotaPozice += B2
                    Case 3
                        v_HodnotaPozice += B3
                    Case 4
                        v_HodnotaPozice += B4
                    Case 5
                        v_HodnotaPozice += B5
                End Select
            ElseIf pocet_plus1 = 0 Then
                Select Case pocet_minus1
                    Case 1
                        v_HodnotaPozice -= B1
                    Case 2
                        v_HodnotaPozice -= B2
                    Case 3
                        v_HodnotaPozice -= B3
                    Case 4
                        v_HodnotaPozice -= B4
                    Case 5
                        v_HodnotaPozice -= B5
                End Select
            End If
        Next

        Return v_HodnotaPozice * HracNaTahu + FieldBonus(PosledniProvedenyTah)

    End Function


    ''' <summary>
    ''' Procedura zajišťující zahrání tahu bez kontoly 5-ky pro porovnávání pozic. 
    ''' </summary>
    ''' <param name="Pole">Array(35) of SByte</param>
    ''' <param name="Tah">Value of Byte</param>
    ''' <param name="Rotace">Value of Byte</param>
    ''' <param name="HracNaTahu">Value of SByte</param>
    ''' <remarks></remarks>
    Private Sub AI_ZahrajTah_SP(ByVal Pole() As SByte, ByVal Tah As Byte, ByVal Rotace As Byte, ByVal HracNaTahu As SByte)

        If Pole(Tah) = 0 Then
            Pole(Tah) = HracNaTahu
        Else
            Alert_Error2()
        End If

        If Rotace >= 4 Then
            If Rotace <= 5 Then
                If Rotace = 4 Then
                    Rotuj_Q3R() '=4
                Else
                    Rotuj_Q3L() '=5
                End If
            Else
                If Rotace = 6 Then
                    Rotuj_Q4R() '=6
                Else
                    Rotuj_Q4L() '=7
                End If
            End If
        Else
            If Rotace >= 2 Then
                If Rotace = 2 Then
                    Rotuj_Q2R() '=2
                Else
                    Rotuj_Q2L() '=3
                End If
            Else
                If Rotace = 0 Then
                    Rotuj_Q1R() '=0
                Else
                    Rotuj_Q1L() '=1
                End If
            End If
        End If

    End Sub


    ''' <summary>
    ''' Procedura zajišťující zahrání tahu. 
    ''' </summary>
    ''' <param name="Tah">Value of Byte</param>
    ''' <param name="Rotace">Value of Byte</param>
    ''' <param name="HracNaTahu">Value of SByte</param>
    ''' <remarks></remarks>
    Private Sub AI_ZahrajTah(ByVal Tah As Byte, ByVal Rotace As Byte, ByVal HracNaTahu As SByte, ByRef INSTANTKILL As Boolean)

        If _Pole(Tah) = 0 Then
            _Pole(Tah) = HracNaTahu
            If AI_PD_Check_For_5_FINAL(_Pole, HracNaTahu) = True Then INSTANTKILL = True
        Else
            Alert_Error2()
        End If

        If Rotace >= 4 Then
            If Rotace <= 5 Then
                If Rotace = 4 Then
                    Rotuj_Q3R() '=4
                Else
                    Rotuj_Q3L() '=5
                End If
            Else
                If Rotace = 6 Then
                    Rotuj_Q4R() '=6
                Else
                    Rotuj_Q4L() '=7
                End If
            End If
        Else
            If Rotace >= 2 Then
                If Rotace = 2 Then
                    Rotuj_Q2R() '=2
                Else
                    Rotuj_Q2L() '=3
                End If
            Else
                If Rotace = 0 Then
                    Rotuj_Q1R() '=0
                Else
                    Rotuj_Q1L() '=1
                End If
            End If
        End If

    End Sub

    ''' <summary>
    ''' Procedura zajištující vrácení zahraného tahu.
    ''' </summary>
    ''' <param name="Tah">Value of Byte</param>
    ''' <param name="Rotace">Value of Byte</param>
    ''' <remarks></remarks>
    Private Sub AI_OdHrajTah(ByVal Tah As Byte, ByVal Rotace As Byte)

        If Rotace >= 4 Then
            If Rotace <= 5 Then
                If Rotace = 4 Then
                    Rotuj_Q3L() '=4
                Else
                    Rotuj_Q3R() '=5
                End If
            Else
                If Rotace = 6 Then
                    Rotuj_Q4L() '=6
                Else
                    Rotuj_Q4R() '=7
                End If
            End If
        Else
            If Rotace >= 2 Then
                If Rotace = 2 Then
                    Rotuj_Q2L() '=2
                Else
                    Rotuj_Q2R() '=3
                End If
            Else
                If Rotace = 0 Then
                    Rotuj_Q1L() '=0
                Else
                    Rotuj_Q1R() '=1
                End If
            End If
        End If

        _Pole(Tah) = 0

    End Sub





#Region "Naivní AI"

    ''' <summary>
    ''' Funkce která vrací nejlepší tah Naivní AI.
    ''' </summary>
    ''' <param name="HracNaTahu">Value of SByte</param>
    ''' <param name="ZapisMysleni">Value of Boolean</param>
    ''' <returns></returns>
    ''' <remarks></remarks>
    Public Function NaivniAI_NejlepsiTah(ByVal HracNaTahu As SByte, ByVal ZapisMysleni As Boolean) As Short

        Dim Tahy(0) As Short
        Dim i, i_best As Integer
        Dim cena, nejlepsi As Single

        Dim PossibleMoves As Short = PocitejMozneTahy()
        ReDim Tahy(PossibleMoves)
        GenerujMozneTahy(Tahy)

        nejlepsi = MinusNekonecno

        i = 0
        Do
            Dim Tah, Rotace As Byte
            Rotace = Tahy(i) \ 36
            Tah = Tahy(i) - 36 * Rotace

            Dim killermove As Boolean = False
            AI_ZahrajTah(Tah, Rotace, HracNaTahu, killermove)
            cena = HodnotaPozice(HracNaTahu, Tah)
            AI_OdHrajTah(Tah, Rotace)

            If (cena > nejlepsi) Then
                i_best = i
                nejlepsi = cena
            End If
            i += 1
        Loop While i < Tahy.Count

        Return Tahy(i_best)

    End Function

#End Region

#Region "Minimax AI"

    ''' <summary>
    ''' Rekurzivní funkce algoritmu Minimax AI.
    ''' </summary>
    ''' <param name="Hloubka">Value of SByte</param>
    ''' <param name="HracNaTahu">Value of SByte</param>
    ''' <param name="PosledniTah">Value of Byte</param>
    ''' <param name="Posledni5Kontrola">Value of SByte</param>
    ''' <param name="HracCekajiciNaVyhru">Value of SByte</param>
    ''' <returns></returns>
    ''' <remarks></remarks>
    Public Function MinMax_Rekurze( _
        ByVal Hloubka As SByte, _
        ByVal HracNaTahu As SByte, _
        ByVal PosledniTah As Byte, _
        ByVal Posledni5Kontrola As SByte, _
        ByVal HracCekajiciNaVyhru As SByte, _
        ByVal killer As Boolean _
        )

        Dim Tahy(0) As Short
        Dim i As Integer
        Dim cena, nejlepsi As Single

        'V případě koncové pozice nebo nulové hloubky propočet ukončíme.


        Dim CheckFor5 As SByte

        If Posledni5Kontrola = 0 Then
            If AI_Check_For_5_FINAL(_Pole) = True Then
                If AI_PD_Check_For_5_FINAL(_Pole, HracNaTahu) = True Then
                    CheckFor5 = HracNaTahu
                    HracCekajiciNaVyhru = HracNaTahu
                End If
            Else
                CheckFor5 = 0
            End If

            If killer = True Then Return MAT_instant
        Else
            If Posledni5Kontrola = 1 Then
                If HracNaTahu = 1 Then
                    Return MAT
                Else
                    If killer = True Then
                        Return 0 'remiza
                    Else
                        Return -MAT
                    End If
                End If
            Else '-1
                If HracNaTahu = -1 Then
                    Return MAT
                Else
                    If killer = True Then
                        Return 0 'remiza
                    Else
                        Return -MAT
                    End If
                End If
            End If
        End If


        'Dávat mat nemůžu, protože pozice je po tahu soupeře.


        If Hloubka <= 0 Then
            Return HodnotaPozice(HracNaTahu, PosledniTah)
            Exit Function
        End If

        'Cyklus přes tahy je podobný jako u předchozího algoritmu.
        Dim PossibleMoves As Short = PocitejMozneTahy()
        ReDim Tahy(PossibleMoves)
        GenerujMozneTahy(Tahy)

        nejlepsi = MinusNekonecno

        i = 0
        Do
            Dim Tah, Rotace As Byte
            Rotace = Tahy(i) \ 36
            Tah = Tahy(i) - 36 * Rotace

            Dim killermove As Boolean = False
            AI_ZahrajTah(Tah, Rotace, HracNaTahu, killermove)
            cena = -MinMax_Rekurze(Hloubka - 1, -HracNaTahu, Tah, CheckFor5, HracCekajiciNaVyhru, killermove) 'zmena -minmaxrekurze -hracnatahu
            AI_OdHrajTah(Tah, Rotace)
            'Nejlepší tah nás nezajímá, důležitá je jen jeho hodnota.
            If (cena > nejlepsi) Then nejlepsi = cena
            i += 1
        Loop While i < Tahy.Count

        'Pokud je výsledná hodnota velmi velká nebo velmi malá, znamená, 
        'že dávám nebo dostávám mat. V tom případě je třeba cenu upravit, 
        'aby se např. z "bílý dává 3. půltahem mat" stalo "černý dostává 
        '4. půltahem mat", neboť se návratem do volající funkce posunu 
        'o jeden půltah. Toto je nezbytné, jinak by program nepreferoval 
        'např. mat 1. tahem před matem 2. tahem.

        'If nejlepsi > MNOHO Then nejlepsi = ReturnMinus(nejlepsi)
        'If nejlepsi < -MNOHO Then nejlepsi = ReturnPlus(nejlepsi)

        Return nejlepsi

    End Function

    ''' <summary>
    ''' Funkce vracející nejlepší tah algoritmu MiniMax AI.
    ''' </summary>
    ''' <param name="Hloubka">Value of SByte</param>
    ''' <param name="HracNaTahu">Value of SByte</param>
    ''' <returns></returns>
    ''' <remarks></remarks>
    Public Function MinMaxAI_NejlepsiTah(ByVal Hloubka As SByte, ByVal HracNaTahu As SByte)

        Dim Tahy(0) As Short
        Dim i, indexNejlepsiho As Integer
        Dim cena, nejlepsi As Single

        'Základní mat
        Dim CheckFor5 As SByte
        If AI_Check_For_5_FINAL(_Pole) = True Then
            If AI_PD_Check_For_5_FINAL(_Pole, 1) = True Then
                CheckFor5 = 1
            Else
                CheckFor5 = -1
            End If
        Else
            CheckFor5 = 0
        End If

        Dim PossibleMoves As Short = PocitejMozneTahy()
        ReDim Tahy(PossibleMoves)
        GenerujMozneTahy(Tahy)

        RedukujTahy(Tahy)

        nejlepsi = MinusNekonecno

        i = 0
        Do
            Dim Tah, Rotace As Byte
            Rotace = Tahy(i) \ 36
            Tah = Tahy(i) - 36 * Rotace

            Dim killermove As Boolean = False
            AI_ZahrajTah(Tah, Rotace, HracNaTahu, killermove)
            cena = -MinMax_Rekurze(Hloubka, -HracNaTahu, Tah, CheckFor5, CheckFor5, killermove) 'zmena -minmax rekurze -hracnatahu
            AI_OdHrajTah(Tah, Rotace)
            If (cena > nejlepsi) Then
                nejlepsi = cena
                indexNejlepsiho = i
            End If
            i += 1
        Loop While i < Tahy.Count

        Return Tahy(indexNejlepsiho)

    End Function

#End Region

#Region "Alfabeta AI"

#Region "Úprava hodnot pozice blízko/daleko od matu"

    ''' <summary>
    ''' Funkce pro úpravu hodnoty při blízkosti matu.
    ''' </summary>
    ''' <param name="Cena">Value of Single</param>
    ''' <returns></returns>
    ''' <remarks></remarks>
    Private Function blizKMatu(ByVal Cena As Single) As Single

        If (Cena > MNOHO) Then
            Cena += 1
        End If
        If (Cena < -MNOHO) Then
            Cena -= 1
        End If

        Return Cena

    End Function

    ''' <summary>
    ''' Funkce pro úpravu hodnoty při vzdálenosti od matu.
    ''' </summary>
    ''' <param name="Cena">Value of Single</param>
    ''' <returns></returns>
    ''' <remarks></remarks>
    Private Function dalOdMatu(ByVal Cena As Single) As Single

        If (Cena > MNOHO) Then
            Cena -= 1
        End If
        If (Cena < -MNOHO) Then
            Cena += 1
        End If

        Return Cena

    End Function

#End Region

    ''' <summary>
    ''' Rekurzivní volání pro funkci AlfaBeta AI.
    ''' </summary>
    ''' <param name="Hloubka">Value of SByte</param>
    ''' <param name="Alfa">Value of Single</param>
    ''' <param name="Beta">Value of Single</param>
    ''' <param name="HracNaTahu">Value of SByte</param>
    ''' <param name="PosledniTah">Value of Byte</param>
    ''' <param name="Posledni5Kontrola">Value of SByte</param>
    ''' <param name="HracCekajiciNaVyhru">Value of SByte</param>
    ''' <returns></returns>
    ''' <remarks></remarks>
    Private Function AlfaBeta_Rekurze( _
        ByVal Hloubka As Byte, _
        ByVal Alfa As Single, _
        ByVal Beta As Single, _
        ByVal HracNaTahu As SByte, _
        ByVal PosledniTah As Byte, _
        ByVal Posledni5Kontrola As SByte, _
        ByVal HracCekajiciNaVyhru As SByte, _
        ByVal killer As Boolean _
    )

        Dim Tahy(0) As Short
        Dim i As Integer
        Dim cena As Single

        'V případě koncové pozice nebo nulové hloubky propočet ukončíme.

        'Základní mat


        Dim CheckFor5 As SByte

        If Posledni5Kontrola = 0 Then
            If AI_Check_For_5_FINAL(_Pole) = True Then
                If AI_PD_Check_For_5_FINAL(_Pole, HracNaTahu) = True Then
                    CheckFor5 = HracNaTahu
                    HracCekajiciNaVyhru = HracNaTahu
                End If
            Else
                CheckFor5 = 0
            End If

            If killer = True Then Return MAT_instant
        Else
            If Posledni5Kontrola = 1 Then
                If HracNaTahu = 1 Then
                    Return MAT
                Else
                    If killer = True Then
                        Return 0 'remiza
                    Else
                        Return -MAT
                    End If
                End If
            Else '-1
                If HracNaTahu = -1 Then
                    Return MAT
                Else
                    If killer = True Then
                        Return 0 'remiza
                    Else
                        Return -MAT
                    End If
                End If
            End If
        End If

        'Dávat mat nemůžu, protože pozice je po tahu soupeře.

        If (Hloubka <= 0) Then Return HodnotaPozice(HracNaTahu, PosledniTah)

        Dim PossibleMoves As Short = PocitejMozneTahy()
        ReDim Tahy(PossibleMoves)
        GenerujMozneTahy(Tahy)

        i = 0
        Do

            Dim Tah, Rotace As Byte
            Rotace = Tahy(i) \ 36
            Tah = Tahy(i) - 36 * Rotace

            Dim killermove As Boolean = False
            AI_ZahrajTah(Tah, Rotace, HracNaTahu, killermove)
            cena = -AlfaBeta_Rekurze(Hloubka - 1, blizKMatu(-Beta), blizKMatu(-Alfa), -HracNaTahu, Tah, CheckFor5, HracCekajiciNaVyhru, killermove)
            cena = dalOdMatu(cena)
            AI_OdHrajTah(Tah, Rotace)
            If (cena > Alfa) Then
                Alfa = cena
                If (cena >= Beta) Then
                    'Zde je právě to ořezání.
                    Return Beta
                End If
            End If
            i += 1
        Loop While i < Tahy.Count
        Return Alfa

    End Function

    ''' <summary>
    ''' Funkce vracející nejlepší tah algoritmu AlfaBeta AI.
    ''' </summary>
    ''' <param name="Hloubka">Value of SByte</param>
    ''' <param name="HracNaTahu">Value of SByte</param>
    ''' <returns></returns>
    ''' <remarks></remarks>
    Public Function AlfaBetaAI_NejlepsiTah(ByVal Hloubka As Byte, ByVal HracNaTahu As SByte)
        Dim Tahy(0) As Short
        Dim i, indexNejlepsiho As Integer
        Dim cena, alfa As Single

        Dim CheckFor5 As SByte
        If AI_Check_For_5_FINAL(_Pole) = True Then
            If AI_PD_Check_For_5_FINAL(_Pole, 1) = True Then
                CheckFor5 = 1
            Else
                CheckFor5 = -1
            End If
        Else
            CheckFor5 = 0
        End If

        Dim PossibleMoves As Short = PocitejMozneTahy()
        ReDim Tahy(PossibleMoves)
        GenerujMozneTahy(Tahy)

        RedukujTahy(Tahy)

        alfa = MinusNekonecno

        i = 0
        Do
            Dim Tah, Rotace As Byte
            Rotace = Tahy(i) \ 36
            Tah = Tahy(i) - 36 * Rotace

            Dim killermove As Boolean = False
            AI_ZahrajTah(Tah, Rotace, HracNaTahu, killermove)
            cena = -AlfaBeta_Rekurze(Hloubka, MinusNekonecno, blizKMatu(-alfa), -HracNaTahu, Tah, CheckFor5, CheckFor5, killermove)
            cena = dalOdMatu(cena)
            AI_OdHrajTah(Tah, Rotace)
            If (cena > alfa) Then
                alfa = cena
                indexNejlepsiho = i
            End If
            i += 1
        Loop While i < Tahy.Count
        Return Tahy(indexNejlepsiho)
    End Function

#End Region













End Class
